body {
display: flex;
flex-direction: column;
height: 100vh;
justify-content: space-evenly;
margin: auto;
text-align: center;
max-width: 42rem;
}

blockquote {
padding: 2rem 0 2rem 0;
}

button, input {
padding: 1.2rem 3rem 1.2rem 3rem;
}

cite {
display: block;
margin: 0.6rem 0 2.2rem 0;
}
